#!/usr/bin/perl

use strict;
use Fcntl qw(O_RDONLY);

use vars qw(%bin
             $dumpDir
             $ptrArgs
             $pwd);

#
# Constants
#
my $kilobyte = 1024;
my $VERSION='1.0.1';
my $defaultraqdomains = 'defaultraqdomains';

my $vsitesDir = "/home/sites";
my $spoolDir = "/var/spool/mail";

#
# parse command line
#
my %arg_opts = ('--help|-h'=>'',
		'--debug|-d'=>'i?',
		'--get-status|-s'=>'',
		'--dump-accounts|-dc'=>'s',
		'--dump-domains|-dd'=>'s',
		'--dump-all|-da'=>'',
		'--get-content-list|-lc'=>'',
		'--no-content|-nc'=>'',
		'--no-compress|-nz'=>'',
		'--output|-o'=>'s',
		'--status-file|-sf'=>'s',
		'--version|-v'=>'',
	       );

%bin = ('tar'=>'tar','bsdtar'=>'bsdtar','split'=>'split','cp'=>'cp','rm'=>'rm','echo'=>'echo',
	   'mysql'=>'mysql','mysqldump'=>'mysqldump',
	   'psql'=>'psql','uptime'=>'uptime','uname'=>'uname',
	   'cat'=>'cat','pwd'=>'pwd','md5sum'=>'md5sum',
	   'grep'=>'grep','find'=>'find','gzip'=>'gzip','wc'=>'wc',
	   'locale'=>'locale', 'iconv'=>'iconv');

$pwd = "";
my $workDir = "";
my $returnCode = 0;

my %UIDB;
my %shadow;


#==============================================================
#
# common subroutines
#

sub printToLog {
  my ($msg, $noeol) = @_;
  if ($msg) {
	print $msg;
  }
  unless ($noeol) {
	print "\n";
  }
}

sub printToError {
  my ($msg, $noeol, $errCode) = @_;
  if ($msg) {
	print STDERR $msg;
  }
  unless ($noeol) {
	print STDERR "\n";
  }
  if (defined($errCode)) {
    $returnCode = $errCode; ## translate an error code to the return code
  } else {
    $returnCode = 63;
  }
}


sub makePasswordNode {
  my $value= shift;
  my $type = shift||'encrypted';

  my $item = &makeXmlNode('password');
  $item->{'CONTENT'}->($value);  ## html encode
  $item->{'ATTRIBUTE'}->('type',$type);
  return $item;
}

sub mangleName {
  my $name = shift;

  if($name =~ /^\d/) {
    $name = 'user' . $name;
  }
  $name =~ s/\./_/g;

  return $name;
}

sub makeSysUserNode {
  my ($name, $password, $type) = @_;

  unless(defined($name) && defined($password) && defined($type)) {
	printToError("makeSysUserNode: Username, password or password type is not set.");
  }

  my $sysUserNode = makeXmlNode('sysuser');
  $sysUserNode->{'ATTRIBUTE'}->('name', mangleName($name));
  $sysUserNode->{'ADDCHILD'}->(makePasswordNode($password, $type));

  return $sysUserNode;
}

# Returns home directory for the system account $user
sub getHomeDir {
	my ($user) = @_;
	my $home;
	my @pw = getpwnam($user);
	if (-1 != $#pw) {
		$home = $pw[7];
	}
	unless ($home) {
		# how it is by default
		$home = '/home/$account';
	}
	return $home;
}

# Returns password for system account $user
sub getSystemPassword {
	my ($user) = @_;
	my $password;
	my @pw = getpwnam($user);
	if (-1 != $#pw) {
		$password = $pw[1];
	}
	unless ($password) {
		$password = '';
	}
	return $password;
}

sub createUserNameFromDomain {
  my ($name) = @_;

  if ($name =~ /^\d/) {
	$name = 'u' . $name;
  }

  $name =~ s/\./_/g;
  $name =~ s/^www\.//g;
  $name =~ s/[aoeui]//g if length($name) > 16;
  $name = substr($name, 0, 16) if length($name) > 16;

  return $name;
}

my $compress = 1;

sub setCompress {
  ($compress) = @_;
}

sub makeDumpDb
{
  my ($dbName, $dbType, $dbUser, $dbPasswd, $dbHost, $ptrDomainBin, $variables, $domainSocket) = @_;
  $dbType ||= 'mysql';

  my ($dumpFile,$dumpSql,$dumpName,$cmd,$gzip,$split,$mysql,$psql, $retCode,$out);

  $dumpName = "$dbName.$dbType.sql";
  $gzip = $compress == 1 ? ($bin{'gzip'} || 'gzip') : undef;

  if (-x $gzip) {
	$dumpName .= ".gz";
  }
  $dumpFile="$dumpDir/$dumpName";

  printToLog(">>Checking database presence");

  if ($dbType eq 'mysql') {
	$mysql = $bin{'mysql'} || 'mysql';
	if (-x $mysql) {
	  $cmd  = "$mysql -u '$dbUser'";
	  $cmd .= " -h $dbHost" if $dbHost;
	  $cmd .= " -p'$dbPasswd'" if $dbPasswd;
	  $cmd .= " -e '' '$dbName' 2>&1";

	  $out = `$cmd`;
	  chomp $out;
	  if ($out) {
		$retCode = $?>>8;
		printToLog("warning: mysql datatbase '$dbName' not found. mysql has returned '$retCode'");
		return "";
	  }
	}
  } elsif ($dbType eq 'Pg') {
	$psql = $bin{'psql'} || 'psql';
	if (-x $psql) {
	  $cmd  = "export PGUSER='$dbUser'";
	  $cmd .= " PGPASSWORD='$dbPasswd'" if $dbPasswd;
	  $cmd .= "; $psql '$dbName' ";
	  $cmd .= " -h $dbHost" if $dbHost;
	  $cmd .= " -c '' 2>&1";

	  $out = `$cmd`;
	  chomp $out;
	  if ($out) {
		$retCode = $?>>8;
		printToLog("warning: PostgreSQL datatbase '$dbName' not found. psql has returned '$retCode'");
		return "";
	  }
	}
  }

  printToLog("..done");

  unless (exists($ptrArgs->{'no-content'})) {
	$gzip = $bin{'gzip'} || 'gzip';
	$split = $bin{'split'} || 'split';

	if ($dbType eq 'mysql') {
	  my ($prefx, $postfix) = ("", "");
	  my $commonprefix = "LD_LIBRARY_PATH=$variables " if $variables;

	  my $commonopts;
	  $commonopts = " --socket='$domainSocket'" if $domainSocket;
	  $commonopts .= " -h $dbHost" if $dbHost;
	  $commonopts .= " -p'$dbPasswd'" if $dbPasswd;
	  $commonopts .= " -u '$dbUser'";

	  my $dumpcmd;
	  if ($ptrDomainBin) {
		$dumpcmd = $ptrDomainBin->{'mysqldump'};
	  } else {
		$dumpcmd = $bin{'mysqldump'} || 'mysqldump';
	  }

	  my $dumpopts = " --quick --quote-names --add-drop-table";

	  my $mysqlversion = `$commonprefix $bin{'mysql'} $commonopts --version`;
	  if ($mysqlversion =~ /Distrib (\d+)\.(\d+)\./) {
		if ($1 > 4 or ($1 == 4 and $2 > 0)) {
		  $dumpopts .= " --compatible=mysql323";
		}
	  }

	  $cmd = $commonprefix . $dumpcmd . $commonopts . $dumpopts . " '$dbName'";

	} elsif ($dbType eq 'Pg') {
	  $cmd  = "export PGUSER='$dbUser'";
	  $cmd .= " PGPASSWORD='$dbPasswd'" if($dbPasswd);
	  $cmd .= ' ; ' . ($bin{'pg_dump'} || 'pg_dump') . ' -d -O';
	  $cmd .= " -h '$dbHost'" if $dbHost;
	  $cmd .= " '$dbName'";
	}

	if ($compress and $gzip and -x $gzip) {
	  $cmd .= " | $gzip";
	}
	if (-x $split) {
	  $cmd .= " | $split -b 2040m - '$dumpFile.'";
	} else {
	  $cmd .= " > '$dumpFile'";
	}

	local $SIG{'PIPE'} = 'DEFAULT' if $SIG{'PIPE'} eq 'IGNORE';

	unless (system($cmd) == 0) {
	  printToError("Error execute '$cmd'");
	  return undef;
	}
  }
  return $dumpName;
}

sub addDbUsers {
  my ($dbNode, $dbName, $wrapUserMysql) = @_;

  printToLog("Dupming $dbName database users");

  $dbName =~ s/_/\\_/g;

  my (@users, $dbUser, $ptrRow);
  my $sql = "SELECT DISTINCT User FROM db WHERE Db = '$dbName'";
  $wrapUserMysql->{'EXECUTE'}->($sql);
  while ($ptrRow = $wrapUserMysql->{'FETCHROW'}->()) {
		push @users, (@{$ptrRow})[0];
  }
  $wrapUserMysql->{'FINISH'}->();

  foreach $dbUser (@users) {
		addDbUser($dbNode, $dbUser, $wrapUserMysql);
  }
}

sub addDbUser {
  my ($dbNode, $user, $wrapUserMysql, $passwordNode, $access_hosts) = @_;

  &printToLog("Dumping mysql user '$user'");

	my ( $item );

  if ( ref( $wrapUserMysql ) =~ /HASH/ &&
			 ref( $wrapUserMysql->{'EXECUTE'} ) =~ /CODE/ ) {

    my ($sql,$ptrRow,@accesshosts,$accesshost);

    $sql = "SELECT password,host FROM user WHERE user='$user'";

    if ( $wrapUserMysql->{'EXECUTE'}->( $sql ) ) {

			$item = makeXmlNode( 'dbuser' );
			$item->{'ATTRIBUTE'}->('name', $user);

      while ( $ptrRow = $wrapUserMysql->{'FETCHROW'}->() ) {
				unless ( ref($passwordNode) =~ /HASH/ ) {
					$passwordNode = makePasswordNode( $ptrRow->[0] );
				}
				push @accesshosts, $ptrRow->[1];
      }

			if ( $passwordNode ) {
				$item->{'ADDCHILD'}->( $passwordNode );
			}

			if (defined($access_hosts) && ref($access_hosts)=~/ARRAY/) {
				@accesshosts = @{$access_hosts};
			}

			foreach $accesshost ( @accesshosts ) {
				$item->{'ADDCHILD'}->( makeXmlNode( 'accesshost', $accesshost ) );
			}

			$dbNode->{'ADDCHILD'}->( $item );

    }

    $wrapUserMysql->{'FINISH'}->();
  }

	return $item;
}

my $lastDumpId = 0;

sub generateDumpId {
  return $$ . "-" . $lastDumpId++;
}

#
# sub makeDumpFile - make a dump of dir tree (use 'tar')
#
# Arguments:
#          $destFile - name of destination file.
#                      If the file has not '.tgz' or '.tar.gz' extension
#                      '.tgz' will be added
#          $scrDir   - path to source dir
#
#          $ptrNames - pointer to an array with names of files and dirs to
#                      include into dump
#                      can be a single name
#                      by default '.' - include all entries
#
#          $ptrExcludes - pointer to an array with names of files and dirs to
#                      exclude from dump
#                      can be a single name
#
#          $extraParams - extra params to run tar
#                         ('-h' - don't dump symlinks; dump the files they point to)
#
sub makeDumpFile {
  my ($destFile, $srcDir, $ptrNames,$ptrExclude,$extraParams)=@_;

  my @destPath = split(/\//, $destFile);
  if (length($destPath[$#destPath]) > 32) {
	$destPath[$#destPath] = generateDumpId();
  }
  $destFile = join('/', @destPath);

  my $tar = $bin{'tar'} || $bin{'bsdtar'} || die "unable to find tar";

  my $split = $bin{'split'}||'split';
  my $gzip = $compress == 1 ? ($bin{'gzip'} || 'gzip') : undef;

  my ($cmd,$ret,@names,@exclude);

  if(ref($ptrNames)=~/ARRAY/){
    @names = @{$ptrNames};
  }else{
    @names=($ptrNames) if $ptrNames;
  }

  if(ref($ptrExclude)=~/ARRAY/){
    @exclude = @{$ptrExclude};
  }else{
    @exclude=($ptrExclude) if $ptrExclude;
  }

  unless(-x $tar ){
    &printToError("Error: '$tar' is not executable");
    return undef;
  }

#
# check extension
#
  if(-x $gzip){
    unless($destFile=~/\.tgz$/ ||$destFile=~/\.tar\.gz$/){
      $destFile .= '.tar.gz';
    }
  }else{
    unless($destFile=~/\.tar$/){
      $destFile .= '.tar';
    }
  }
#
# end check extension
#

  unless( exists $ptrArgs->{'no-content'}){

    $cmd=$tar;
    if($srcDir){
      if( -d $srcDir){
		$cmd.=" -C $srcDir -f - -c ";
		if ($bin{'tar'}) { # GNU tar present -> understand --ignore-failed-read
		  $cmd .= " --ignore-failed-read ";
		}
      }else{
		return undef;
      }
    }


    if($extraParams){
      $cmd .= $extraParams.' ';
    }

    if(@exclude){
      $cmd .= ' '.join( ' ', map {s/^\Q$srcDir\E\///;'--exclude '.$_} @exclude); ## extract domain name
    }

	# stop treating arguments as options, if tar is recent enough
	if ($bin{'tar-accepts-doubledash'}) {
	  $cmd .= ' -- ';
	}

    if(@names){
	  @names = map { "'$_'" } @names;
      if($srcDir){
		$cmd .= join( ' ', map {s/^\Q$srcDir\E\///;$_} @names); ## extract domain name
      }else{
		$cmd .= join( ' ', @names );
      }
    }else{
      $cmd .= '.';  ## dump all files & dirs
    }

    if(-x $gzip){
      $cmd .= " | $gzip";
    }
    if(-x $split){
      $cmd .= " | $split -b 2040m - $destFile.";
    }else{
      $cmd .= " >$destFile";
    }

		local $SIG{'PIPE'} = 'DEFAULT' if $SIG{'PIPE'} eq 'IGNORE';

    unless(system($cmd) == 0){
      &printToError("Error execute command: '$cmd': $!");
      $destFile='';
    }
  }

  $destFile=~s/^\Q$dumpDir\E\///;  ### remove dir name
  return $destFile;
}

sub getDumpDir {
  my ($subdir,$clean) = @_;
  my ($dumpdir);

  if($subdir=~/^\//){
    $dumpdir=$subdir;
  }else{
    my $util = $bin{'pwd'}||'pwd';
    $dumpdir = $pwd||`$util`;
    chomp $dumpdir;
    if($subdir){
      $dumpdir.='/'.$subdir;
    }
  }
  if(-d $dumpdir && $clean){
    unless(system("rm -rf $dumpdir")==0){
      &printToError("Error: can not clean temporary dir '$dumpdir': $!");
    }
  }
  unless(-d $dumpdir){
    mkdir ($dumpdir,0755) or
      die "Error: can not create temporary dir '$dumpdir': $!\n";
  }
  return $dumpdir;
}

sub printXml {
  my ($root,$outFh,$dtd) = @_;
  unless($outFh){
    $outFh = *STDOUT;
  }

  unless(ref($root)=~/HASH/ && ref($root->{'PRINT'})=~/CODE/){
    print $outFh "Error: a xml root is not found\n";
    &printToError("Error: sub printXml: bad arguments");
    return 0;
  }
  print $outFh "<?xml version='1.0' encoding=\"UTF-8\"?>\n";

  my  $doctype = $root->{'NAME'}->();
  if($dtd){
    print $outFh "<!DOCTYPE $doctype SYSTEM '$dtd'>\n";
  }
  $root->{'PRINT'}->($outFh,'  ');
  return 1;
}

#
# XmlNode("blah", "content" => "blah",
#                 "children" => [$foo, $bar],
#                 "attributes" => {"botva"=>"7", "foo"=>"bar"})
#
sub XmlNode {
  my ($name, %args) = @_;

  my $node;
  if ($args{'content'}) {
	$node = makeXmlNode($name, $args{'content'});
  } else {
	$node = makeXmlNode($name);
  }

  if ($args{'attributes'}) {
	foreach my $attribute (keys %{$args{'attributes'}}) {
	  $node->{'ATTRIBUTE'}->($attribute, $args{'attributes'}->{$attribute});
	}
  }

  if ($args{'children'}) {
	foreach my $child (@{$args{'children'}}) {
	  $node->{'ADDCHILD'}->($child);
	}
  }
  return $node;
}

sub makeXmlNode {
  my ($this,$parent,%attributes,@content);
  my ($name,$plain)=@_;

  $plain = toUtf8(encodeHtml($plain));

  unless($name||$plain){
    return undef;
  }
  $this = {
	   'ATTRIBUTE' => sub {
	     my $key = shift;
	     if(@_){
	       my $value=shift;
	       if(@_){
					 $attributes{$key} = $value;
	       }else{
					 $attributes{$key} = toUtf8(normalizeAttribute($value));
	       }
	     }
	     return $attributes{$key};
	   },
	   'PARENT' => sub {
	     if(@_){
	       my $arg = shift;
	       if(ref($arg)=~/HASH/){
					 $parent=shift;
	       }
	     }
	     return $parent;
	   },
	   'NAME' => sub {
	     if(@_){
	       $name = shift;
	     }
	     return $name;
	   },
	   'CONTENT' => sub {
	     if(@_){
	       $plain = toUtf8(encodeHtml(shift));
	     }
	     return $plain;
	   },
	   'ADDCHILD'=> sub {
	     my $kind = shift;
	     if(
					(ref($kind)=~/HASH/) &&
					(ref($kind->{'PARENT'})=~/CODE/)
	       ) {
	       push @content,$kind;
	       $kind->{'PARENT'}->($this);
	     }
	     return $#content;
	   },
	   'KIND'=> sub {
	     my $index = shift;
			 my ($ret);
			 if ( $index =~ /^\d+$/ ) {
				 if($index >= 0 && $index <= $#content ) {
					 $ret =  $content[$index];
				 }
			 } else {
				 foreach my $item (@content){
					 if ( $item->{'NAME'}->() eq $index ) {
						 $ret = $item;
						 last;
					 }
				 }
			 }
			 return $ret;
	   },
	   'KINDSREF'=> sub {
			 my $name = shift;
			 my $ret = \@content;
			 if ( $name ) {
				 $ret = [];
				 foreach my $item (@content) {
					 if ( $item->{'NAME'}->() eq $name ) {
						 push @{$ret}, $item;
					 }
				 }
			 }
	     return $ret;
	   },
	   'ATTRIBUTESREF' => sub {
	     return \%attributes;
	   },
	   'PRINT'=>sub {
	     my $fh = shift || *STDOUT;
	     my $addPrefix = shift || '';
	     my $prefix = shift || '';
	     if($name){
	       my ($out,$item);
	       if(@content || defined($plain)){

					 $out = $prefix . &getTag($name,\%attributes,'start');
					 print $fh $out;
					 if(@content){
						 print $fh "\n";
						 foreach $item (@content){
							 if(ref($item)=~/HASH/ && ref($item->{'PRINT'})=~/CODE/){
								 $item->{'PRINT'}->($fh,$addPrefix,$prefix.$addPrefix);
							 }
						 }
					 }
					 if(defined($plain)){
						 print $fh $plain;
						 if(@content){
							 print $fh "\n$prefix";
						 }
					 }else{
						 if($prefix){
							 print $fh $prefix;
						 }
					 }
					 $out = &getTag($name,\%attributes,'stop') . "\n";
					 print $fh $out;

	       }else{
					 $out = $prefix . &getTag($name,\%attributes) . "\n";
					 print $fh $out;
	       }
	     }else{
	       if(defined($plain)){
					 print $fh $plain;
	       }
	     }
	   }
	 };

  return $this;
}

sub getTag {
  my ($name,$ptrAttr,$mode) = @_;
  my ($out,$key,$value);
  $mode||='';
  if ($mode eq 'stop'){
    $out="</$name>";
  }else{
    $out="<$name";
    while (($key,$value)=each(%{$ptrAttr})){
      $out .= ' '.$key;
      if (defined($value)){
		$out .= '="'.$value.'"';
      }
    }
    if($mode eq 'start'){
      $out .= ">";
    }else{
      $out .= "/>";
    }
  }
  return $out;
}

#
# end class for handle with XML
#


#
# class for handle with database
#
sub getDbConnect {
  my $ptrBin = shift;
  my $myDbType = shift;
  my $myDbAddress = shift ;
  my $myDbUser = shift ;
  my $myDbPasswd = shift ;
  my $myDbDB = shift ;
  my $myDbHost = shift ;
  my $continueAfterError = shift;

  #added by Honig for Sphera
  my $myDbSocket = shift;
  my $variables = shift;

  my ($dbh,$this,$sth);
  my ($shell,$args);
  my $echo = $ptrBin->{'echo'}||'echo';
  my ($sql,@result,$resultIndex,$ptrResult);

  unless(defined($myDbPasswd) && defined($myDbUser)){
    if($myDbType eq 'mysql'){
      foreach my $myCnf ('/etc/my.cnf',$ENV{'HOME'}.'/.my.cnf'){
		next unless (-T $myCnf);
		next unless open(CNF,"<$myCnf");
		binmode CNF;
		my($key,$value);
		my $state=0;
		while (<CNF>){
		  next unless $_;
		  next if (/^#/);
		  if($state){
			if(/^\s*\[/){
			  last;
			}
			if(/^\s*(\S+)\s*=\s*(\S+)/){
			  $key=$1;
			  $value=$2;
			  if($key =~ /^user$/){
				$myDbUser = $value;
				# sometimes strings are quoted
				$myDbUser =~ s/^('|")(.*)\1$/$2/;
			  }elsif($key =~ /^pass$/){
				$myDbPasswd = $value;
				# sometimes strings are quoted
				$myDbPasswd =~ s/^('|")(.*)\1$/$2/;
			  }
			}
		  }else{
			if(/^\s*\[client\]/){
			  $state=1;
			}
		  }
		}
		close (CNF);
      }
    }elsif($myDbType eq 'Pg'){
      $myDbPasswd ||= $ENV{'PGPASSWORD'};
      $myDbUser ||= $ENV{'PGUSER'};
    }
  }

  if (eval "require DBI") {
    if (eval "require DBD::$myDbType"){
      if($myDbAddress){
		if($myDbUser && $myDbPasswd){
		  unless ($dbh = DBI->connect($myDbAddress,$myDbUser,$myDbPasswd)) {
			return if $continueAfterError;
			die "Error: unable to connect to database: $DBI::errstr\n";
		  }
		}elsif($myDbUser){
		  unless ($dbh = DBI->connect($myDbAddress,$myDbUser)) {
			return if $continueAfterError;
			die "Error: unable to connect to database: $DBI::errstr\n";
		  }
		}else{
		  unless ($dbh = DBI->connect($myDbAddress)) {
			return if $continueAfterError;
			die "Error: unable to connect to database: $DBI::errstr\n";
		  }
		}
      }
    }
  }

#  unless($dbh){
#    if ($myDbType eq 'Pg'){
#      if(eval "require Pg"){
#        $myDbAddress="user=$myDbUser dbname=$myDbDB";
#        if($myDbPasswd){
#          $myDbAddress.=" password=$myDbPasswd";
#        }
#        $dbh = Pg::connectdb($myDbAddress);
#      }
#    }
# }
	unless ($dbh) {

		if ($myDbType eq 'mysql') {
			$shell = $ptrBin->{'mysql'} || 'mysql';
			$args  = " --socket='$myDbSocket'" if ($myDbSocket);
			if ($myDbPasswd) {
				$args .= " -B -u '$myDbUser' -p'$myDbPasswd' '$myDbDB' ";
			} else {
				$args .= " -B -u $myDbUser $myDbDB ";
			}
			$shell = "LD_LIBRARY_PATH=$variables $shell" if ($variables);
    }elsif($myDbType eq 'Pg'){
      $shell = $ptrBin->{'psql'}||'psql';
	  if($myDbPasswd) {
		$shell = "PGPASSWORD=$myDbPasswd $shell";
	  }
	  if($myDbUser) {
		$shell = "PGUSER=$myDbUser $shell";
	  }

      $args = " -d $myDbDB -c ";
    }
#
# check connection to Confixx databnase
#
    if($myDbDB=~/confixx/){
      $sql = "select hostname from admin";
      @result = `$echo "$sql"| $shell $args`;
      unless($result[0]=~/^\w+$/){
		$args = ' -h $myDbHost'.$args;
		@result = `$echo "$sql"| $shell $args`;
		unless($result[0]=~/^\s*hostname/){
		  die "Error: Can not connect to Confixx database (unable to execute sql: $sql)\n@result\n";
		}
      }
    }
#
# end check connection to Confixx databnase
#
    $shell .= $args;
  }

  $this = {
		'RECONNECT'=>sub{
			if($dbh){
				$dbh->disconnect();
				if($myDbPasswd){
					$dbh = DBI->connect($myDbAddress,$myDbUser,$myDbPasswd) or
						die "Error: unable to connect to database: $DBI::errstr\n";
				}else{
					$dbh = DBI->connect($myDbAddress,$myDbUser) or
						die "Error: unable to connect to database: $DBI::errstr\n";
				}
				return $dbh;
			}else{
				return 1;
			}
		},
	   'EXECUTE_INT' => sub {
	     my $sql = shift;
		 my $already_reconnected = shift;
	     if ($dbh) {
	       $sth = executeDbh($dbh,$sql);
	       if ($sth) {
			 return $sth->rows;
	       } else {
			 # Wired MySQL timeouts on some DBI installations even on actively used connections
			 if ($DBI::err == 2006 && !$already_reconnected) {
				printToError("MySQL connection timeout detected - attempting to reconnect... ", 1);
				unless ($this->{'RECONNECT'}->()) {
					printToError("FAILED");
					return 0;
				}
				printToError("OK");
				return $this->{'EXECUTE_INT'}->($sql, 1);
			 }
			 return 0;
	       }
	     } else {
	       $resultIndex = 0;
	       executeDbShell($shell, \@result, $echo, $sql, $myDbType);
	       return scalar(@result);
	     }
	   },
	   'EXECUTE' => sub {
		 if ($sql ne "") {
		   print "---> Unfinished statement $sql\n";
		 }
	     $sql = shift;
		 my $res = $this->{'EXECUTE_INT'}->($sql);
		 unless ($res) {
		   $sql = "";
		 }
		 return $res;
	   },
	   'FETCHROW'=> sub {
	     if($sth){
	       if(@result = $sth->fetchrow){
					 return \@result;
	       }else{
					 return undef;
	       }
	     }else{
	       $resultIndex++;
	       if($resultIndex>$#result){
					 return undef;
	       }else{
					 return $result[$resultIndex];
	       }
	     }
	   },
	   'FETCHHASH'=>sub {
	     if($sth){
	       $ptrResult = $sth->fetchrow_hashref();
	     }else{
	       $resultIndex++;
	       if($resultIndex>$#result){
					 $ptrResult = undef;
	       }else{
					 my ($i,$key);
					 $ptrResult = {};
					
					 for($i=0; $i<@{$result[0]}; $i++){
						 $ptrResult->{$result[0]->[$i]}=$result[$resultIndex]->[$i];
					 }
	       }
	     }
	     return $ptrResult;
	   },
		 'FIELDS' => sub { ## get list of field names
			 if($sth){
				 my @fields = ();
				 for(my $i=0; $i<$sth->{'NUM_OF_FIELDS'}; $i++){
					 push @fields, $sth->{'NAME'}->[$i];
				 }
				 return @fields;
			 }else{
				 return @{$result[0]};
			 }					
		 },
	   'FINISH'=>sub {
	     if($sth){
	       $sth->finish();
	     }
		 $sql = "";
	     @result=();
	     $ptrResult = undef;
	   },
	   'PRINTERROR'=>sub {
	     if($dbh){
	       if(@_){
			 $dbh->{'PrintError'}=shift;
	       }
	       return $dbh->{'PrintError'};
	     }else{
	       return 0;
	     }
	   },
	   'PING'=>sub{
	     if($dbh){
	       return $dbh->ping();
	     }else{
	       return 1;
	     }
	   },
	   'DISCONNECT'=>sub{
	     if($dbh){
	       return $dbh->disconnect();
	     }else{
	       return 1;
	     }
	   },
	   'USER'=>sub{
	     return $myDbUser;
	   },
	   'PASSWORD'=>sub{
	     return $myDbPasswd;
	   },
	   'HOST'=>sub{
	     return $myDbHost;
	   },
	   'TYPE'=>sub{
	     return $myDbType;
	   },
	  };

  return $this;
}

sub executeDbh {
  my $dbh = shift;
  unless(ref($dbh)=~/DBI/){
    &printToError ("Error: DBI object is not found");
  }
  my $sql = shift;

  my $sth = $dbh->prepare($sql) or
    &printToError ("Error: prepare sql: $sql\n'" . $DBI::errstr. "'", 0);
  if($sth->execute){
    return $sth;
  }else{
    &printToError ("Error: execute sql: $sql\n'" . $DBI::errstr."'", 0);
  }
  return undef;
}

sub executeDbShell {
  my ($shell,$ptrRes,$echo,$sql, $type) = @_;
  @{$ptrRes}=();
  my (@out,@row,$ret);
  $ret=0;
  if($type eq 'Pg'){
    @out = `$shell "$sql"`;
  }else{
	# backticks are used to quote `ssl` for MySQL 4
	$sql =~ s/`/\\`/g;
    @out = `$echo "$sql"|$shell`;
  }
  if ($out[0]=~/^error/i){
    &printToErrror ("Error: execute sql: $sql\n@out");

  }else{

    if($type eq 'Pg'){
      foreach my $line (@out){
				chomp $line;
				next if ($line=~/^-[-+]*$/); # skip delimiter
				if($line=~/^\(\s*\d+\s+\w+\)$/){
					last;  ## (6 rows)
				}
				$line=~s/^\s+//;
				$line=~s/\s+$//;
				@row = split(/\s*\|\s*/,$line);
				push @{$ptrRes}, [@row];
				$ret++;
			}

    }else{
      if($out[0]=~/^\+[-+]*$/){
				foreach my $line (@out){
					chomp $line;
					next if ($line=~/^\+[-+]*$/); # skip delimiter
					$line=~s/^\|\s*//;
					$line=~s/\s+$//;
					@row = split(/\s*\|\s*/,$line);
					@row = map { $_ eq 'NULL' ? undef : $_ } @row;
					push @{$ptrRes}, [@row];
					$ret++;
				}
      }else{  ## batch mode
				foreach my $line (@out){
					chomp $line;
					@row = split(/\t/,$line);
					@row = map { $_ eq 'NULL' ? undef : $_ } @row;
					push @{$ptrRes}, [@row];
					$ret++;
				}
      }
    }
  }
  return $ret;
}

#
# end class for handle with database
#

sub makeMIMEBase64 {
  my ($useModule,$this);
  if (eval "require MIME::Base64"){
    $useModule = 1;
  }else{
    $useModule = 0;
  }

  $this={
	 'ENCODE'=>sub{
	   my $src = shift;
	   my ($encoded);
	   if($useModule){
	     $encoded = MIME::Base64::encode($src);
		 chomp $encoded;
	   }else{
	     $encoded = &encodeBase64($src);
	   }
	   return $encoded;
	 },
	 'USEMODULE'=>sub{
	   return $useModule;
	 },
	};

  return $this;
}

#
# url decode (certificates in Plesk DB encoded by the url encoding)
#

sub urlDecode {
    my $url = $_[0];
    $url =~ tr/+/ /;
    $url =~ s/%([a-fA-F0-9]{2,2})/chr(hex($1))/eg;
    return $url;
}

sub encodeHtml {
  my(@ret,$src);
  if ($src=shift){
    if(ref $src){
      if(${$src}=~/[&"'><]/){
		${$src}=~s/&/&amp;/sg;
		${$src}=~s/>/&gt;/sg;
		${$src}=~s/</&lt;/sg;
		${$src}=~s/\'/&#39;/sg;
		${$src}=~s/\"/&quot;/sg;
      }
      return ${$src};
    }else{
      if($src=~/[&"'><]/){
		$src=~s/&/&amp;/sg;
		$src=~s/>/&gt;/sg;
		$src=~s/</&lt;/sg;
		$src=~s/\'/&#39;/sg;
		$src=~s/\"/&quot;/sg;
      }
    }
  }
  return $src;
}

# Normalization according to the XML 1.0 (Third Edition),
# clauses 2.3 "Common Syntactic Constructs"
# and 3.3.3 "Attribute-value normalization"

sub normalizeAttribute {
  $_ = $_[0];
  s/\&/&amp;/g;
  s/\r/&#xD;/g;
  s/\n/&#xA;/g;
  s/\t/&#x9;/g;
  s/\</&lt;/g;
  s/\"/&quot;/g;
  s/\'/&#39;/g;
  return $_;
}

############################
#
#	Non-ascii characters handling
{
	# static
	my $charset;

	sub setCharset {
	  $charset = shift;
	}

	# try to determine the local charmap
	sub _determineCharset()
	{
	  my $charset;
	  my $locale = $bin{'locale'} || 'locale';

	  if (exists $ENV{'LANG'}) {
		my @pieces = split /\./, $ENV{'LANG'};
		if (defined $pieces[1]) {
		  $charset = $pieces[1];
		  chomp $charset;
		}
	  }
	  if (not $charset && open(CM, "$locale charmap |")) {
	  	binmode CM;
		$charset = <CM>;
		close(CM);
		chomp $charset;
	  }
	  if (not $charset) {
		$charset = 'ISO-8859-1';
	  }

	  return $charset;
	}

	sub toUtf8
	{
		my ($subj, $myCharset) = @_;

		$charset = _determineCharset() if (not $charset);

		$myCharset = $charset unless ( $myCharset );

		if ( (not $subj) ||
				 (not $bin{'iconv'}) ||
				 ( $myCharset =~ /UTF-?8/i )
			 ) {
			return $subj;
		}

		if ($subj =~ /^[\x00-\x7f]*$/) { return $subj; }

		my $converted;
		my $escaped = $subj;
		$escaped =~ s/'/'"'"'/g;

		if (open(IC, "echo '$escaped' | " . $bin{'iconv'} . " -f '$myCharset' -t 'UTF-8' |")) {
			binmode IC;
			while (my $string = <IC>) { $converted .= $string; }
			close(IC);
			chomp $converted;
			if ($converted) {
			  return $converted;
			}
		}

		return "";
	}
}

#
# Based on the b64.c by Bob Trower (base64.sf.net)
#

my @cb64 = ('A', 'B', 'C', 'D', 'E', 'F', 'G', 'H',
			'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P',
			'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X',
			'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f',
			'g', 'h', 'i', 'j',	'k', 'l', 'm', 'n',
			'o', 'p', 'q', 'r', 's', 't', 'u', 'v',
			'w', 'x', 'y', 'z', '0', '1', '2', '3',
			'4', '5', '6', '7', '8', '9', '+', '/');

@cb64 = map { ord } @cb64;

sub encodeBlock {
  my ($len, @in) = @_;
  my @res;
  push @res, $cb64[$in[0] >> 2];
  push @res, $cb64[(($in[0] & 0x3)<<4)|(($in[1]&0xf0)>>4)];
  push @res, $len>1 ? $cb64[(($in[1]&0xf)<<2)|(($in[2]&0xc0)>>6)] : ord('=');
  push @res, $len>2 ? $cb64[$in[2]&0x3f] : ord('=');
  return @res;
}

sub encodeBase64 {
  my @in = unpack "C*", $_[0];

  my @out;
  while (scalar(@in)) {
	my @block = (0, 0, 0);
	my $length = 0;
	while (scalar(@in) && $length < 3) {
	  $block[$length++] = shift(@in);
	}
	push @out, encodeBlock($length, @block);
  }
  return pack "C*", @out;
}

sub resolveFile {
  my ($prefixes, $file) = @_;

  foreach my $prefix (@{$prefixes}) {
	$prefix.='/' unless $prefix=~/\/$/;
	$file =~ s|^/||;

	return ($prefix.$file) if (-x ($prefix.$file));
  }
}

sub checkGnuTar {
  my ($file) = @_;

  # on some FreeBSD this command outputs on stderr
  open (TAR, "$file --version 2>&1 |");
  while (<TAR>) {
	return 1 if (/GNU tar/);
  }
  return 0;
}

sub getBins {
  my ($ptrBin,$path) = @_;
  my (%searchPath,$key,$prefix,$bin,$fullPath);
  my ($notFound,$i,@prefixes);

  $notFound=0;

  $i=0;
  foreach $prefix (split(/:/,$path)){
    next unless $prefix;
    unless(exists($searchPath{$prefix})){
      $searchPath{$prefix}=$i++;
      push @prefixes,$prefix;
    }
  }

  foreach $key (keys %{$ptrBin}){
	$ptrBin->{$key} = resolveFile(\@prefixes, $ptrBin->{$key});

	# special case: 'tar' *must* be GNU tar

	if ($key eq 'tar') {
	  my $foundGnuTar = checkGnuTar($ptrBin->{'tar'});

	  # if 'tar' is not GNU tar, try 'gtar'
	  if (!$foundGnuTar) {
		if (checkGnuTar(resolveFile(\@prefixes, 'gtar'))) {
		  $ptrBin->{'tar'} = resolveFile(\@prefixes, 'gtar');
		} else {
		  $ptrBin->{'tar'} = undef;
		}
	  }

	  if ($ptrBin->{'tar'}) {
		open (TAR, "$ptrBin->{'tar'} --version|");
		while (<TAR>) {
		  if (/tar \(GNU tar\) (\d+)\.(\d+)/) {
			if ($1 > 1 or ($1 == 1 and $2 >= 13)) {
			  $ptrBin->{'tar-accepts-doubledash'} = 1;
			}
			last;
		  }
		}
		close (TAR);
	  }
	}

	# bsdtar case: it is fallback case and should be set up
	# 1) to 'tar' if 'tar' is not GNU tar
	# 2) to 'bsdtar' if no 'tar' found in system

	if ($key eq 'bsdtar') {
	  # if 'bsdtar' not found, but 'tar' exists and not a GNU tar
	  unless ($ptrBin->{'bsdtar'}) {
		if (resolveFile(\@prefixes, 'tar')) {
		  unless (checkGnuTar(resolveFile(\@prefixes, 'tar'))) {
			$ptrBin->{'bsdtar'} = resolveFile(\@prefixes, 'bsdtar');
		  }
		}
	  }
	}

    $notFound++;
  }

  return $notFound;
}

sub getArguments {
  my ($ptrArgv,$ptrOpts) = @_;

  my (@keys,$firstKey,%prepKeys,$key,$value,$ptrArr,$arg,$state);
  my (%retHash,$pat,$found,$used,@rest,$fullArg,$prevKey);

  while (($key,$value)=each(%{$ptrOpts})){
    @keys = split(/\|/,$key);
    $firstKey = $keys[0];
    $firstKey =~s/^-*//;
    push @{$prepKeys{$firstKey}},[@keys];
    push @{$prepKeys{$firstKey}},$value;
  }

  $state =0;
  foreach $arg (@{$ptrArgv}){

#    print "arg: $arg\n";

    $used = 0;
    if($state==1){
      if  ($arg =~ /^-/){
	$state=0;
      }else{
	$used = 1;
	$retHash{$prevKey}=$arg;
	next;
      }
    }
    if ($state == 2){
      $retHash{$prevKey}=$arg;
      $state = 0;
      $used = 1;
    }else{
      $fullArg = $arg;
      if ($arg =~ /^(-\S+)=(\S+)/){
	$arg = $1;
	$value = $2;
      }else{
	$value = undef;
      }
      foreach $key (keys %prepKeys){

	$ptrArr = $prepKeys{$key};

#	print "key: $key, ",join('|',@{$ptrArr->[0]}),"\n";


	$found = 0;
	foreach $pat (@{$ptrArr->[0]}){
	  if ($pat eq $arg){
	    $found = 1;
	    last;
	  }
	}
	if($found){
	  $used = 1;
	  if(defined($value)){
	    $retHash{$key}=$value;
	  }else{
	    if($ptrArr->[1]){
	      if($ptrArr->[1] =~ /\?$/){
		$state = 1;
	      }else{
		$state = 2;
	      }
	      $prevKey = $key;
	    }else{
	      $retHash{$key}=undef;
	    }
	  }
	  last;
	}
      }
    }
    unless($used){
      if($value){
	push @rest,$fullArg;
      }else{
	push @rest,$arg;
      }
    }
  }
  @{$ptrArgv}=@rest;
  return \%retHash;

}

sub makeDumpStatus($) {
  my $statusFile = shift;
  my($pid,$currAccount,$currDomain,$leftAccounts,$leftDomains,$this);
  $pid=$$;
  $this = {'FILE'=>sub {
	     if(@_){
	       $statusFile = shift;
	     }
	     return $statusFile;
	   },
	   'PID'=>sub {
	     if(@_){
	       $pid = shift;
	     }
	     return $pid;
	   },
	   'ACCOUNT'=>sub{
	     if(@_){
	       $currAccount = shift;
	       if($leftAccounts){
		 $leftAccounts--;
	       }
	     }
	     return $currAccount;
	   },
	   'DOMAIN'=>sub{
	     if(@_){
	       $currDomain = shift;
	       if($leftDomains){
		 $leftDomains--;
	       }
	     }
	     return $currDomain;
	   },
	   'COUNTS'=>sub{
	     if(@_){
	       my $newValue = shift;
	       if (defined $newValue){
		 $leftAccounts=$newValue;
	       }
	       $newValue = shift;
	       if (defined $newValue){
		 $leftDomains=$newValue;
	       }
	     }
	     return ($leftAccounts,$leftDomains);
	   },
	   'PRINT'=>sub{
	     if(@_){
	       $statusFile=shift;
	     }
	     open(STATUS,">$statusFile.tmp");

		 binmode STATUS;
	     my $root = &makeXmlNode('agent-dump-status');
	     my $item = &makeXmlNode('pid',$pid);
	     $root->{'ADDCHILD'}->($item);
	     $item = &makeXmlNode('account',$currAccount||'');
	     $root->{'ADDCHILD'}->($item);
	     $item = &makeXmlNode('domain',$currDomain||'');
	     $root->{'ADDCHILD'}->($item);
	     $item = &makeXmlNode('accounts-left',
				  defined($leftAccounts)?$leftAccounts:'');
	     $root->{'ADDCHILD'}->($item);
	     $item = &makeXmlNode('domains-left',
				  defined($leftDomains)?$leftDomains:'');
	     $root->{'ADDCHILD'}->($item);

	     &printXml($root,*STATUS,'agent-dump-status.dtd');

	     close(STATUS);

		 rename("$statusFile.tmp", $statusFile);
	   }

	  };
  return $this;
}

sub makeContentList {
  my ($dirDump,$xmlDump) = @_;
  my ($name,$fullPath,%files,%fromXML,$size,@cids,$root,$item,$key,$cmd,
      $gzip,$wc,$md5sum);

  if(opendir(DUMP,$dirDump)){
    while ($name=readdir(DUMP)){
      $fullPath="$dirDump/$name";
      unless(-f $fullPath && $name=~/\.\w\w$/){
	next;
      }
      $key = $fullPath;
      $key =~ s/\.\w\w$//;
      $size=(stat($fullPath))[7];
      if(defined $size){
	push @{$files{$key}},[$fullPath,$size];
      }
    }
    closedir(DUMP);
  }

	if (-T $xmlDump) {
		if (open(DUMP, "<$xmlDump")) {
			binmode DUMP;
			while (<DUMP>) {
				if (@cids = /cid(?:[_-]\w+)?=\"([^"]+)\"/g) {
					foreach $fullPath (@cids) {
						unless ($fullPath =~ /^\//) {
							$fullPath = "$dirDump/$fullPath";
						}
						$fromXML{$fullPath} = undef;
					}
				}
			}
			close(DUMP);
		}

		foreach $key (keys %files) {
			unless (exists($fromXML{$key})) {
				delete $files{$key};
			}
		}
	}

  $gzip = $bin{'gzip'}||'gzip';
  $wc = $bin{'wc'}||'wc';
  $md5sum = $bin{'md5sum'}||'md5sum';

  $root=&makeXmlNode('contentlist');

  my($compressed,$uncompressed,$list,$count,$ptrList,@sortedList,
     $ptrItem,$out);
  my($totalUncomp,$totalSize)=(0,0);
  foreach $key (keys %files){

#
# sort by extension  .aa .ab ...
#
    @sortedList =  sort {&getExt($a->[0]) cmp &getExt($b->[0])} @{$files{$key}};

#
# output list of files
#
    $compressed=0;
    foreach $ptrItem (@sortedList){
      ($fullPath,$size)=@{$ptrItem};
      $item=&makeXmlNode('file',$fullPath);
      $item->{'ATTRIBUTE'}->('size',$size,'noencode');
      $root->{'ADDCHILD'}->($item);

#
# get md5sum
#
      if(-x $md5sum){
	$out=`$md5sum $fullPath`;
	if($out=~/\s*([0-9a-fA-F]+)\s+/){
	  $item->{'ATTRIBUTE'}->('md5sum',$1,'noencode');
	}
      }

      $compressed += $size;

    }

    $uncompressed = undef;
#
# get unpacked size of the archive
#
    if ($key=~/\.tgz$|\.gz$/){
      if(-x $gzip){
	if(scalar(@sortedList)==1){
	  ($fullPath,$size)=@{$sortedList[0]};
	  ($uncompressed,$compressed)=&getZipSize($fullPath,$gzip);
	}else{
	  $list = join(' ',map {$_->[0]} @sortedList);
	  if(-x $wc){
	    $cmd = "cat $list | $gzip | $wc -c";
	    $uncompressed = `$cmd`;
	    chomp $uncompressed;
	    if ($uncompressed =~ /^\s*(\d+)\s*$/){
	      $uncompressed = $1;
	    }else{
	      $uncompressed = undef;
	    }
	  }else{
	    $cmd = "cat $list >$key";
	    if(system($cmd)==0){
	      ($uncompressed,$compressed)=&getZipSize($key,$gzip);
	      unlink $key;
	    }
	  }
	}
      }
    }
#
# calc total size
#
    $totalSize += $compressed;
    $totalUncomp += $uncompressed if $uncompressed;
  }

  $root->{'ATTRIBUTE'}->('size',$totalSize,'noencode');
  if ($totalUncomp && $totalUncomp != $totalSize){
    $root->{'ATTRIBUTE'}->('unpacked',$totalUncomp,'noencode');
  }

  return $root;
}

sub getExt {
  my $path = shift;
  if($path=~/\.([^.]*)$/){
    return $1;
  }
  return $path;
}

sub getZipSize {
  my ($fullPath,$gzip) = @_;
  $gzip ||= 'gzip';
  my $cmd = "$gzip -l $fullPath";
  my ($uncompressed,$compressed,@out);
  if (open(OUTPUT,"$cmd|")){
    ($compressed,$uncompressed)=(0,0);
    while(<OUTPUT>){
      @out = split(' ',$_);
      if($out[0]=~/^\d+$/ && $out[1]=~/^\d+$/){
	$uncompressed+=$out[1];
	$compressed+=$out[0];
      }
    }
    close (OUTPUT);
  }
  return $uncompressed,$compressed;
}

my $outputIsFile;

sub openOutput
{
	my ($outPath) = @_;
	my ($outFh);

	if ($outPath) {
		open(OUTPUT, ">>$outPath") or die "Error: can not open '$outPath' for write: $!\n";
		binmode OUTPUT;
		$outFh = *OUTPUT;
		$outputIsFile = 1;
	} else {
		$outFh = *STDOUT;
		$outputIsFile = 0;
	}

	return $outFh;
}

sub closeOutput{
  if($outputIsFile){
    close (OUTPUT) or
      &printToError("Error: unable to close OUTPUT: $!\n");
    $outputIsFile=0;
  }
}

#=========================================================================
#
# dtd-specific (but common) subroutines
#
sub createPasswordNode {
  my ($parent, $password, $type) = @_;

  my $item = makeXmlNode('password', $password);
  $item->{'ATTRIBUTE'}->('type', $type);
  $parent->{'ADDCHILD'}->($item);
}

sub createSysuserNode {
  my ($parent, $name, $password, $type) = @_;

  my $item = makeXmlNode('sysuser');
  $item->{'ATTRIBUTE'}->('name', $name);
  createPasswordNode($item, $password, $type);
  $parent->{'ADDCHILD'}->($item);
}

sub readFile {
  my ($fileName) = @_;

  unless (-f $fileName) {
	return undef;
  }

  open(INPUT, "< $fileName");
  binmode INPUT;
  my $content = join('', <INPUT>);
  close(INPUT);

  return $content;
}

1;

#
# end common subroutines
#
#=========================================================================


$ptrArgs = getArguments(\@ARGV, \%arg_opts);

#
# search binaries
#
getBins(\%bin,'/bin:/usr/bin:/usr/local/bin:/sbin:/usr/sbin:/usr/local/sbin:'.$ENV{'PATH'});

my ($outPath,$outFh,$outHas,$dumpFile,$statusFile,$app);

$app = $0;
if ($app =~ /(.+)\/([^\/]+)$/) {
  $workDir = $1;
  $app = $2;
}
$app=~s/\.pl$//;

$pwd=`$bin{'pwd'}`||'.';
chop $pwd;

if ($workDir eq '.') {
  $workDir = $pwd;
} elsif ($workDir !~ /^\//) {
  $workDir = $pwd . '/' . $workDir;
}

$dumpFile = $workDir . '/RaQ2.xml';
$statusFile = $workDir . '/RaQ2.status';
$dumpDir = $workDir . '/pma';

my $objDumpStatus = &makeDumpStatus($ptrArgs->{'status-file'}||$statusFile);

if(!$ptrArgs->{'output'} &&
   (exists $ptrArgs->{'dump-all'} ||
    exists $ptrArgs->{'dump-accounts'} ||
    exists $ptrArgs->{'dump-domains'}))
{
  $ptrArgs->{'output'} = $dumpFile;
}

if ($outPath = $ptrArgs->{'output'}) {
	open(OUTPUT, ">$outPath") or die "Error: can not open '$outPath' for write: $!\n";
	binmode OUTPUT;
	$dumpFile = $outPath;
	close(OUTPUT);
}

if (exists $ptrArgs->{'version'}) {
  print $outFh $VERSION,"\n";
  exit 0;
}

#
# get MIME Base64 encoder
#
my $wrapBase64 = makeMIMEBase64();

my $version = getVersion();
if ($version) {
  eval "require Cobalt::Vsite";
  eval "require Cobalt::Stats";
  eval "require Cobalt::Group";
  eval "require Cobalt::Fpx";
  eval "require Cobalt::Ftp";
  eval "require Cobalt::List";
  eval "require Cobalt::User";
  eval "require Cobalt::Vacation";
  eval "require Cobalt::Email";
}

if (exists $ptrArgs->{'get-status'}) {
  my $outFh = openOutput($ptrArgs->{'output'});
  if ($version) {
	printAgentStatus($ptrArgs->{'output'} || $dumpFile, $objDumpStatus->{'FILE'}->(), $outFh);
  } else {
	printWrongPlatform($outFh);
  }
  closeOutput();
  exit(0);
}

if(exists ($ptrArgs->{'dump-all'}) ||
   exists ($ptrArgs->{'dump-accounts'}) ||
   exists ($ptrArgs->{'dump-domains'})) {

  my ($root, @accounts, @domains, $ptrAccounts, $ptrDomains, $value);
  if (exists $ptrArgs->{'no-content'}) {
    $::dumpDir = getDumpDir("$workDir/pma");
  } else {
    $::dumpDir = getDumpDir("$workDir/pma", 'clean');
  }

  if (exists $ptrArgs->{'no-compress'}) {
	setCompress(0);
  }

  printToLog("Work dir: $workDir");
  printToLog("Dump file: $dumpFile");
  printToLog("Status file: " . $objDumpStatus->{'FILE'}->());

  if ($value = $ptrArgs->{'dump-accounts'}) {
  	if ($value eq "-") {
	  $value = <STDIN>;
	  chomp $value;
	}
    @accounts = split(/\s*,\s*/, $value);
    $ptrAccounts = \@accounts;
  }

  if ($value = $ptrArgs->{'dump-domains'}) {
  	if ($value eq "-") {
	  $value = <STDIN>;
	  chomp $value;
	}
    @domains = split(/\s*,\s*/,$value);
    $ptrDomains = \@domains;
  }
#
# generate a xml dump
#
  $root = getRaQ2Dump($ptrAccounts, $ptrDomains);
#
# print dump to output
#
  printToLog('Print dump file ...', 1);
  if (ref($root) =~ /HASH/ && ref($root->{'PRINT'}) =~ /CODE/) {
    my $outFh = openOutput($ptrArgs->{'output'});
    printXml($root, $outFh, 'RaQ2.dtd');
    closeOutput();
  }
  printToLog(' OK');
} elsif (exists $ptrArgs->{'get-content-list'}) {
  $::dumpDir = getDumpDir("$workDir/pma");
  my $root = makeContentList($::dumpDir, $dumpFile);
  if (ref($root) =~ /HASH/) {
    my $outFh = openOutput($ptrArgs->{'output'});
    printXml($root, $outFh);
    closeOutput();
  }
} else {
  printHelp();
}

exit $returnCode;

#
# end main
#
#==============================================================
#
# subroutines
#

sub openUIDB {
  tie(%UIDB, 'DB_File', '/var/cobalt/uidb', Fcntl::O_RDONLY, 0600);
}

sub closeUIDB {
  untie(%UIDB);
}

sub parseShadow {
  open SHADOW, "/etc/shadow";
  while (<SHADOW>) {
	my @userInfo = split /:/;
	if ($userInfo[1] ne '*') {
	  $shadow{$userInfo[0]} = $userInfo[1];
	}
  }
  close SHADOW;
}

sub getRaQ2Dump {
  my ($ptrAccounts, $ptrDomains) = @_;

  openUIDB();
  parseShadow();

  my $root = makeXmlNode('RaQ2dump');
  my $client = dumpClient();

  if (ref($ptrDomains) =~ /ARRAY/) {
	dumpDomains($client, $ptrDomains);
  } else {
	dumpAll($client);
  }

  $root->{'ADDCHILD'}->($client);

  closeUIDB();

  return $root;
}

sub dumpClient {
  my $clientNode = makeXmlNode('client');
  $clientNode->{'ATTRIBUTE'}->('name', 'raqdomains');
  $clientNode->{'ATTRIBUTE'}->('contact', 'RaQ domains');
  createPasswordNode($clientNode, '', 'empty');

  return $clientNode;
}

sub dumpAll {
  my ($parent) = @_;

  my @domains = map {$_->[1]} Cobalt::Vsite::vsite_list();
  dumpDomains($parent, \@domains);
}

sub dumpDomains {
  my ($parent, $ptrDomains) = @_;

  my (%domainsInfo, $domain);
  foreach $domain (Cobalt::Vsite::vsite_list()) {
	$domainsInfo{$domain->[1]} = $domain;
  }

  foreach $domain (@{$ptrDomains}) {
	if (exists $domainsInfo{$domain}) {
	  dumpDomain($parent, $domainsInfo{$domain});
	} elsif (exists $domainsInfo{'www.'.$domain}) {
	  dumpDomain($parent, $domainsInfo{'www.'.$domain});
	}
  }
}

sub dumpDomain {
  my ($parent, $domainInfo) = @_;

  my $www = 'false';
  if ($domainInfo->[1] =~ /^www./) {
	$domainInfo->[1] =~ s/^www.//;
	$www = 'true';
  }

  my $domainNode = makeXmlNode('domain');
  $parent->{'ADDCHILD'}->($domainNode);

  $domainNode->{'ATTRIBUTE'}->('ip', $domainInfo->[0]);
  $domainNode->{'ATTRIBUTE'}->('name', $domainInfo->[1]);
  $domainNode->{'ATTRIBUTE'}->('www', $www);

  # NO MORE INFO REQUIRED IN '--no-content' MODE
  if (exists $ptrArgs->{'no-content'}) {
	return;
  }

  my $domainHandle = $domainInfo->[2];

  my $stat_period = Cobalt::Stats::stats_get_www_report($domainHandle);
  if ($stat_period eq 'weekly' or $stat_period eq 'daily') {
	$domainNode->{'ATTRIBUTE'}->('stat-period', $stat_period);
  }

  if (-d "$vsitesDir/$domainHandle/web") {
	my $dump = makeDumpFile("$dumpDir/$domainHandle.web", "$vsitesDir/$domainHandle/web");
	if ($dump) {
	  $domainNode->{'ATTRIBUTE'}->('cid', $dump);
	}
  }

  dumpDomainLimits($domainNode, $domainHandle);
  dumpDomainPermissions($domainNode, $domainHandle);
  dumpAnonFtp($domainNode, $domainHandle);
  dumpUsers($domainNode, $domainInfo->[1], $domainHandle);
  dumpMailLists($domainNode, $domainInfo->[1], $domainHandle);
}

sub dumpDomainLimits {
  my ($parent, $domainHandle) = @_;

  my $disk_quota = (Cobalt::Group::group_get_quota($domainHandle))[1];
  if ($disk_quota && $disk_quota != 0) {
	my $limitNode = makeXmlNode('domain-limit', $disk_quota*$kilobyte);
	$limitNode->{'ATTRIBUTE'}->('name', 'quota');
	$parent->{'ADDCHILD'}->($limitNode);
  }

  my $user_quota = $UIDB{'maxUsers-' . $domainHandle};
  if ($user_quota && $user_quota != 0) {
	my $limitNode = makeXmlNode('domain-limit', $user_quota);
	$limitNode->{'ATTRIBUTE'}->('name', 'users');
	$parent->{'ADDCHILD'}->($limitNode);
  }
}

sub dumpDomainPermissions {
  my ($parent, $domainHandle) = @_;

  createPermNode($parent, 'domain', 'email', $UIDB{'emailDomain-' . $domainHandle} eq 'on');
  createPermNode($parent, 'domain', 'shell', $UIDB{'shell-' . $domainHandle} eq 'on');
  createPermNode($parent, 'domain', 'cgi', Cobalt::Vsite::vsite_get_cgis($domainHandle));
  createPermNode($parent, 'domain', 'ssi', Cobalt::Vsite::vsite_get_ssi($domainHandle));
  createPermNode($parent, 'domain', 'frontpage', Cobalt::Fpx::fpx_get_web($domainHandle));
}

sub dumpAnonFtp {
  my ($parent, $domainHandle) = @_;

  my @anonftp = Cobalt::Ftp::ftp_get_anonymous($domainHandle);

  if ($anonftp[0] eq '1') {
	my $anonFtpNode = makeXmlNode('anonftp');

	my $usersLimit = $anonftp[1] || $UIDB{'maxFtpUsers-' . $domainHandle} || 10;
	my $usersLimitNode = makeXmlNode('anonftp-limit', $usersLimit);
	$usersLimitNode->{'ATTRIBUTE'}->('name', 'users');
	$anonFtpNode->{'ADDCHILD'}->($usersLimitNode);

	my $incomingQuota = $anonftp[2] || 20;
	my $incomingQuotaNode = makeXmlNode('anonftp-limit', $incomingQuota*$kilobyte);
	$incomingQuotaNode->{'ATTRIBUTE'}->('name', 'incoming-quota');
	$anonFtpNode->{'ADDCHILD'}->($incomingQuotaNode);

	if (-d "$vsitesDir/$domainHandle/ftp/incoming") {
	  my $dump = makeDumpFile("$dumpDir/$domainHandle.ftp.incoming", "$vsitesDir/$domainHandle/ftp/incoming");
	  if ($dump) {
		$anonFtpNode->{'ATTRIBUTE'}->('cid', $dump);
	  }
	}
	if (-d "$vsitesDir/$domainHandle/ftp") {
	  my $dump = makeDumpFile("$dumpDir/$domainHandle.ftp", "$vsitesDir/$domainHandle/ftp", ".", "incoming");
	  if ($dump) {
		$anonFtpNode->{'ATTRIBUTE'}->('cid_incoming', $dump);
	  }
	}
	$parent->{'ADDCHILD'}->($anonFtpNode);
  }
}

sub dumpUsers {
  my ($parent, $domainName, $domainHandle) = @_;

  my @users = Cobalt::List::list_get_mem($domainHandle . '-users');

  my @admins = Cobalt::Vsite::vsite_admin_list($domainHandle);
  my @admins = sort @admins;
  my $adminName = @admins[0] if @admins > 0;

  my $user;
  foreach $user (@users) {
	dumpUser($parent, $domainHandle, $domainName, $user, $user eq $adminName);
  }

  unless ($adminName) {
	dumpSyntheticAdmin($parent, $domainHandle, $domainName);
  }
}

sub dumpUser {
  my ($parent, $domainHandle, $domainName, $user, $admin) = @_;

  my ($name, $fullname, $quota, , $shell) = Cobalt::User::user_show($user);

  my $userNode = makeXmlNode($admin ? 'admin' : 'user');

  if ($fullname) {
	$userNode->{'ATTRIBUTE'}->('fullname', $fullname);
  }

  $userNode->{'ADDCHILD'}->(makeSysUserNode($name, $shadow{$user}, 'encrypted'));

  my $quotaNode = makeXmlNode('user-limit', $quota*$kilobyte);
  $quotaNode->{'ATTRIBUTE'}->('name', 'quota');
  $userNode->{'ADDCHILD'}->($quotaNode);

  createPermNode($userNode, 'user', 'shell', $shell eq '/bin/bash');
  createPermNode($userNode, 'user', 'admin', $admin);
  createPermNode($userNode, 'user', 'frontpage', Cobalt::Fpx::fpx_get_web($user) == 1);

  if (-d "$vsitesDir/$domainHandle/users/$name/web") {
	my $dump = makeDumpFile("$dumpDir/$name"."@"."$domainHandle.web",
							"$vsitesDir/$domainHandle/users/$name/web");
	if ($dump) {
	  $userNode->{'ATTRIBUTE'}->('cid', $dump);
	}
  }

  dumpMail($userNode, $domainName, $domainHandle, $parent, $user);

  $parent->{'ADDCHILD'}->($userNode);
}

sub dumpSyntheticAdmin {
  my ($parent, $domainHandle, $domainName) = @_;

  my $name = createUserNameFromDomain($domainName);

  my $adminNode = makeXmlNode('admin');
  $adminNode->{'ATTRIBUTE'}->('fullname', $domainName . " admin");
  $adminNode->{'ADDCHILD'}->(makeSysUserNode($name, '', 'empty'));

  my $adminMailNode = makeXmlNode('mail');
  $adminMailNode->{'ATTRIBUTE'}->('mailname', $name);

  $adminNode->{'ADDCHILD'}->($adminMailNode);

  $parent->{'ADDCHILD'}->($adminNode);
}

sub dumpMail {
  my ($parent, $domainName, $domainHandle, $domainNode, $user) = @_;

  my $mailNode = makeXmlNode('mail');
  $mailNode->{'ATTRIBUTE'}->('mailname', $user);

  my @forwards = Cobalt::List::alias_get_vacationless($user);
  my $forward;
  foreach $forward (@forwards) {
	$mailNode->{'ADDCHILD'}->(makeXmlNode('forward', $forward));
  }

  my @aliases = Cobalt::Email::mail_virtuser_get_byuser($user);
  @aliases = map { s/(.+)\@.*$/$1/; $_; } @aliases;
  @aliases = grep !/^$user$/, @aliases;

  my ($alias, %uniquealiases);
  foreach $alias (@aliases) {
	$alias =~ s/^\./_/gi;
	$alias =~ s/\.$/_/gi;
	$uniquealiases{$alias} = 1;
  }

  foreach $alias (keys %uniquealiases) {
	if ($alias eq '@' . $domainName) {
	  $domainNode->{'ATTRIBUTE'}->('catch-all', $user . '@' . $domainName);
	} elsif ($alias =~ /@/) {
	  printToLog("Unable to parse mailalias '$alias', skipping");
	} else {
	  $mailNode->{'ADDCHILD'}->(makeXmlNode('alias', $alias));
	}
  }

  my $msg = $wrapBase64->{'ENCODE'}->(Cobalt::Vacation::vacation_get_message($user));
  my $autoresponderNode = makeXmlNode('autoresponder', $msg);
  my $autoresponderState = Cobalt::Vacation::vacation_get_on($user);
  $autoresponderNode->{'ATTRIBUTE'}->('status', $autoresponderState == '1' ? 'on' : 'off');

  if ($msg || $autoresponderState) {
	$mailNode->{'ADDCHILD'}->($autoresponderNode);
  }

  if (-f "$spoolDir/$user") {
	my $dump = makeDumpFile("$dumpDir/$user"."@"."$domainHandle.mail", $spoolDir, $user);
	if ($dump) {
	  $mailNode->{'ATTRIBUTE'}->('cid', $dump);
	}
  }

  $parent->{'ADDCHILD'}->($mailNode);
}

sub dumpMailLists {
  my ($parent, $domainName, $domainHandle) = @_;

  my @lists = Cobalt::List::list_sites_lists($domainHandle);

  my $list;
  foreach $list (@lists) {
	dumpMailList($parent, $domainHandle, $domainName, $list);
  }
}

sub dumpMailList {
  my ($parent, $domainHandle, $domainName, $list) = @_;

  my @members = Cobalt::List::list_get_mem($list . '_' . $domainHandle);

  my $mailListNode = makeXmlNode('maillist');
  $mailListNode->{'ATTRIBUTE'}->('name', $list . '_' . $domainHandle);

  my $member;
  foreach $member (@members) {
	if ($member =~ /@/) {
	  $mailListNode->{'ADDCHILD'}->(makeXmlNode('recipient', $member));
	} else {
	  $mailListNode->{'ADDCHILD'}->(makeXmlNode('recipient', $member . '@' . $domainName));
	}
  }

  $parent->{'ADDCHILD'}->($mailListNode);
}

sub createPermNode {
  my ($parent, $prefix, $name, $value) = @_;

  if ($value) {
	my $permNode = makeXmlNode($prefix . '-permission');
	$permNode->{'ATTRIBUTE'}->('name', $name);
	$parent->{'ADDCHILD'}->($permNode);
  }
}

sub printAgentStatus {
  my ($dumpFile, $statusFile, $outFh) = @_;
  my $outFh ||= *STDOUT;

  my ($root, $out, $item);

  $root = &makeXmlNode('agentstatus');

  $root->{'ADDCHILD'}->(makeXmlNode('dumpfile', $dumpFile));
  $root->{'ADDCHILD'}->(makeXmlNode('statusfile', $statusFile));

  printXml($root, $outFh, 'RaQ2.dtd');
}

sub printWrongPlatform {
  my ($outFh) = @_;

  my $root = makeXmlNode('agentstatus');
  my $item = makeXmlNode('wrong-platform',
					   $bin{'iconv'} ? '' : 'no iconv found on the source host');
  $root->{'ADDCHILD'}->($item);
  printXml($root, $outFh, 'agentstatus.dtd');
  return;
}

sub getVersion {
  my %bldHash = ('2700'=>'RaQ1',
				 '2799'=>'RaQ2',
				 '2800'=>'RaQ2',
				 '3000'=>'RaQ3',
				 '3001'=>'RaQ4',
				 '3100'=>'RaQ4',
				 '3500'=>'RaQXTR',
				 '3599'=>'RaQXTR',
				);

  my ($build, $version);

  if ((-T '/etc/build') && open (CONF, "</etc/build")) {
    while (<CONF>) {
      if (/(\d{4,4})/) {
		my $buildCode = $1;
		if (exists($bldHash{$1})) {
		  $build = $1;
		  $version = $bldHash{$1};
		  last;
		}
      }
    }
    close (CONF);
  }

  if ($version eq 'RaQ2') {
	return "RaQ2";
  }
}

sub printHelp {

  print <<"HELP";

Agent of migration from 'RaQ2' to 'Plesk 7'

Usage:
  $0 <options>

Options:
  -s |--get-status           get status of the agent
  -dc|--dump-accounts=<list> a coma separated list of resellers to dump
  -dd|--dump-domains=<list>  a coma separated list of domains   to dump
  -da|--dump-all             make a full dump

  -lc|--get-content-list     get list of content files
  -nc|--no-content           do not make content files
  -nz|--no-compress          do not compress content files

  -o |--output=<path>        redirect output to file

  -h |--help                 this help
  -d |--debug=<level>        level of debug messages
  -v |--version

HELP
}
